@@ -1,9 +1,9 @@
 #  You may distribute under the terms of either the GNU General Public License
 #  or the Artistic License (the same terms as Perl itself)
 #
-#  (C) Paul Evans, 2021 -- leonerd@leonerd.org.uk
+#  (C) Paul Evans, 2021-2022 -- leonerd@leonerd.org.uk
 
-package XS::Parse::Keyword 0.21;
+package XS::Parse::Keyword 0.33;
 
 use v5.14;
 use warnings;
@@ -277,12 +277,84 @@ restored again when parsing has finished.
 
 =head2 XPK_ANONSUB
 
-I<atomic, emits op.>
+I<atomic, emits cv.>
 
 A brace-delimited block of code is expected, and assembled into the body of a
 new anonymous subroutine. This will be passed as a protosub CV in the I<cv>
 field.
 
+=head2 XPK_STAGED_ANONSUB
+
+   XPK_STAGED_ANONSUB(stages ...)
+
+I<structural, emits cv.>
+
+A variant of C<XPK_ANONSUB> which accepts additional function pointers to be
+invoked at various points during parsing and compilation. These can be used to
+interrupt the normal parsing in a manner similar to L<XS::Parse::Sublike>,
+though currently somewhat less flexibly.
+
+The I<stages> list may contain elements of the following types. Not every
+stage must be present, but any that are present must be in the following
+order. Multiple copies of each stage are permitted; they are invoked in the
+written order, with parser code happening inbetween.
+
+=over 4
+
+=item XPK_ANONSUB_PREPARE
+
+   XPK_ANONSUB_PREPARE(&callback)
+
+I<atomic, emits nothing.>
+
+Invokes the callback before C<start_subparse()>.
+
+=item XPK_ANONSUB_START
+
+   XPK_ANONSUB_START(&callback)
+
+I<atomic, emits nothing.>
+
+Invokes the callback after C<block_start()> but before parsing the actual
+block contents.
+
+=item XPK_ANONSUB_END
+
+   OP *op_wrapper_callback(pTHX_ OP *o, void *hookdata);
+
+   XPK_ANONSUB_END(&op_wrapper_callback)
+
+I<atomic, emits nothing.>
+
+Invokes the callback after parsing the block contents but before calling
+C<block_end()>. The callback may modify the optree if required and return a
+new one.
+
+=item XPK_ANONSUB_WRAP
+
+   XPK_ANONSUB_WRAP(&op_wrapper_callback)
+
+I<atomic, emits nothing.>
+
+Invokes the callback after C<block_end()> but before passing the optree to
+C<newATTRSUB()>. The callback may modify the optree if required and return a
+new one.
+
+=back
+
+=head2 XPK_ARITHEXPR
+
+I<atomic, emits op.>
+
+   XPK_ARITHEXPR
+
+An arithmetic expression is expected, parsed using C<parse_arithexpr()>, and
+passed as an optree in the I<op> field.
+
+=head2 XPK_ARITHEXPR_VOIDCTX, XPK_ARITHEXPR_SCALARCTX
+
+Variants of C<XPK_ARITHEXPR> which puts the expression in void or scalar context.
+
 =head2 XPK_TERMEXPR
 
 I<atomic, emits op.>
@@ -296,6 +368,18 @@ an optree in the I<op> field.
 
 Variants of C<XPK_TERMEXPR> which puts the expression in void or scalar context.
 
+=head2 XPK_PREFIXED_TERMEXPR_ENTERLEAVE
+
+   XPK_PREFIXED_TERMEXPR_ENTERLEAVE(pieces ...)
+
+A variant of C<XPK_TERMEXPR> which expects a sequence pieces first before it
+parses a term expression, similar to how C<XPK_PREFIXED_BLOCK_ENTERLEAVE>
+works. The entire operation is wrapped in an C<ENTER>/C<LEAVE> pair.
+
+This is intended just for use of C<XPK_SETUP> pieces as prefixes. Any other
+pieces which actually parse real input are likely to cause overly-complex,
+subtle, or outright ambiguous grammars, and should be avoided.
+
 =head2 XPK_LISTEXPR
 
 I<atomic, emits op.>
@@ -376,6 +460,19 @@ character. It is passed as a L<version> SV object in the I<sv> field.
 The C<_OPT>-suffixed version is optional; if no version string is found then
 I<sv> is set to C<NULL>.
 
+=head2 XPK_LEXVAR
+
+I<atomic, emits padix.>
+
+   XPK_LEXVAR(kind)
+
+A lexical variable name is expected and looked up from the current pad. The
+resulting pad index is passed in the I<padix> field. No error happens if the
+variable is not found; the value C<NOT_IN_PAD> is passed instead.
+
+The C<kind> argument specifies what kinds of variable are permitted, as per
+C<XPK_LEXVARNAME>.
+
 =head2 XPK_LEXVAR_MY
 
 I<atomic, emits padix.>
@@ -394,6 +491,20 @@ I<atomic, can probe, emits nothing.>
 
 A literal character (C<,>, C<:> or C<=>) is expected. No argument value is passed.
 
+=head2 XPK_AUTOSEMI
+
+I<atomic, emits nothing.>
+
+A literal semicolon (C<;>) as a statement terminator is optionally expected.
+If the next token is a closing brace to indicate the end of a block, then a
+semicolon is not required. If anything else is encountered an error will be
+raised.
+
+This piece type is the same as specifying the C<XPK_FLAG_AUTOSEMI>. It is
+useful to put at the end of a sequence that forms part of a choice of syntax,
+where some forms indicate a statement ending in a semicolon, whereas others
+may end in a full block that does not need one.
+
 =head2 XPK_INFIX_*
 
 I<atomic, can probe, emits infix.>
@@ -424,8 +535,19 @@ Generally it is best reserved just for the first component of a
 C<XPK_OPTIONAL> or C<XPK_REPEATED> sequence, to provide a "secondary keyword"
 that such a repeated item can look out for.
 
-This was previously called C<XPK_STRING>, and is provided as a synonym for
-back-compatibility but new code should use this new name instead.
+=head2 XPK_KEYWORD
+
+I<atomic, can probe, emits nothing.>
+
+   XPK_KEYWORD("keyword")
+
+A literal string match is expected. No argument value is passed.
+
+This is similar to C<XPK_LITERAL> except that it additionally checks that the
+following character is not an identifier character. This ensures that the
+expected keyword-like behaviour is preserved. For example, given the input
+C<"keyword">, the piece C<XPK_LITERAL("key")> would match it, whereas
+C<XPK_KEYWORD("key")> would not because of the subsequent C<"w"> character.
 
 =head2 XPK_SEQUENCE
 
@@ -525,6 +647,20 @@ I<structural, can probe, emits nothing.>
 A structural type which expects to find a sequence of pieces, all contained in
 parentheses as C<( ... )>. This will pass no extra arguments.
 
+=head2 XPK_ARGSCOPE
+
+I<structural, emits nothing.>
+
+   XPK_ARGSCOPE(pieces ...)
+
+A structural type similar to C<XPK_PARENSCOPE>, except that the parentheses
+themselves are optional; much like Perl's parsing of calls to known functions.
+
+If parentheses are encountered in the input, they will be consumed by this
+piece and it will behave identically to C<XPK_PARENSCOPE>. If there is no open
+parenthesis, this piece will behave like C<XPK_SEQUENCE> and consume all the
+pieces inside it, without expecting a closing parenthesis.
+
 =head2 XPK_BRACKETSCOPE
 
 I<structural, can probe, emits nothing.>
@@ -577,6 +713,20 @@ from the scope itself.
 This is a convenient shortcut to nesting the scope within a C<XPK_OPTIONAL>
 macro.
 
+=head2 XPK_..._pieces
+
+   XPK_SEQUENCE_pieces(ptr)
+   XPK_OPTIONAL_pieces(ptr)
+   ...
+
+For each of the C<XPK_...> macros that takes a variable-length list of pieces,
+there is a variant whose name ends with C<..._pieces>, taking a single pointer
+argument directly. This must point at a C<const XSParseKeywordPieceType []>
+array whose final element is the zero element.
+
+Normally hand-written C code of a fixed grammar would be unlikely to use these
+forms, but they may be useful in dynamically-generated cases.
+
 =cut
 
 =head1 AUTHOR
